import numpy as np
import matplotlib.pyplot as plt
import cv2
from scipy.ndimage import gaussian_filter1d, median_filter
from filterpy.kalman import KalmanFilter

def fspecial_gaussian_1d(size, sigma):
    kernel = np.zeros(size)
    kernel[size//2] = 1
    kernel = gaussian_filter1d(kernel, sigma)
    return kernel/np.sum(kernel)

def lowpass_filtering(data, alpha, T, sample_rate):

    print(f'Low pass filtering camera path with alpha={alpha} and T={T}...')

    N = len(data)

    fft_data = np.fft.fft(data)
    dt = 1 / sample_rate
    # Frequency axis
    filtered_fft_data = np.zeros_like(fft_data)
    freq = np.fft.fftfreq(N, dt)

    # Apply raised cosine filter in frequency domain
    raised_cosine_filter = np.sinc(freq / (2 * T)) * np.cos(np.pi * alpha * freq / T) / (1 - (2 * alpha * freq / T)**2)
    filtered_fft_data[:,0] = np.real(fft_data[:,0]) * raised_cosine_filter
    filtered_fft_data[:,1] = np.real(fft_data[:,1]) * raised_cosine_filter
    # Perform Inverse FFT
    filtered_data = np.real(np.fft.ifft(filtered_fft_data))

    return filtered_data

def gaussian_smoothing(data, sigma):

    print(f'Gaussian smoothing camera path with sigma={sigma}...')

    # Generate 1D Gaussian kernel
    
    filtSize = int(data.shape[0] * (((5*sigma) + 1)/630))

    gaussian_kernel = fspecial_gaussian_1d(filtSize, sigma) 

    # Perform convolution with the Gaussian kernel
    out = np.ones_like(data)
    out[:,0] = np.convolve(data[:,0], gaussian_kernel, mode='same')
    out[:,1] = np.convolve(data[:,1], gaussian_kernel, mode='same')
    out[:,2] = np.convolve(data[:,2], gaussian_kernel, mode='same')
    return out

# didn't work as well 
def mean_filtering(data, filtSize):

    print(f'Mean filtering camera path with filter_size={filtSize}...')
    scaled_filtSize = int(data.shape[0] * (filtSize/630))
    """Perform average filtering on 1D data."""
    smoothed_data = np.ones_like(data)
    mean_kernel = np.ones(scaled_filtSize) / scaled_filtSize
    smoothed_data[:,0] = np.convolve(data[:,0], mean_kernel, mode='same')
    smoothed_data[:,1] = np.convolve(data[:,1], mean_kernel, mode='same')
    smoothed_data[:,2] = np.convolve(data[:,2], mean_kernel, mode='same')
    return smoothed_data

# didn't work as well 
def median_filtering(data, filtSize):

    print(f'Median filtering camera path with filter_size={filtSize}...')

    smoothed_data = np.ones_like(data)
    smoothed_data[:,0] = median_filter(data[:,0], size=filtSize)
    smoothed_data[:,1] = median_filter(data[:,1], size=filtSize)
    smoothed_data[:,2] = median_filter(data[:,1], size=filtSize)
    return smoothed_data

def kalman_filtering(data, N, process_variance, measurement_variance):

    print(f'Kalman filtering camera path with measurement_variance={measurement_variance}...')

    measurements = data[:,:2]
    # Create a Kalman filter with a 2D state (position in x and y)
    kf = KalmanFilter(dim_x=2, dim_z=2)

    # Initialize state variables (position in x and y)
    kf.x = np.array([measurements[0, 0], measurements[0, 1]])

    # Define the state transition matrix
    kf.F = np.array([[1, 0],
                     [0, 1]])

    # Define the measurement matrix
    kf.H = np.array([[1, 0],
                     [0, 1]])

    # Set the initial covariance matrix
    kf.P *= 1e2

    # Set process and measurement noise
    kf.Q = np.array([[process_variance, 0],
                     [0, process_variance]])
    kf.R = np.array([[measurement_variance, 0],
                     [0, measurement_variance]])

    # Kalman filtering
    filtered_positions = np.ones_like(data)
    for i in range(N):
        kf.predict()
        kf.update(measurements[i])
        filtered_positions[i,:2] = kf.x.copy()

    return filtered_positions
