import numpy as np

from optimization_tools.l1_optimization import stabilize, write_output

from pipeline_tools.frame_generation import generate_frames
from pipeline_tools.feature_extraction import get_feature_transform_list, get_feature_transform_list_opt
from pipeline_tools.calculate_trajectory import get_camera_trajectory
from pipeline_tools.visualization import graph_paths, graph_comparison_filtered_paths
from pipeline_tools.stabilization import get_stabilized_transforms_and_trajectory
from pipeline_tools.video_reconstruction import frames_to_video, reconstruct_new_video

from filtering_tools.filtering_functions import lowpass_filtering, gaussian_smoothing, mean_filtering, kalman_filtering

def filtered_methods_ideal(video_name, keyword): 

    frames = generate_frames(video_name)
    N = len(frames)

    original_transforms, frame_shape = get_feature_transform_list(frames, N)

    original_path = get_camera_trajectory(original_transforms)

    paths = {}

    # Lowpass filter
    sample_rate = 30.0  # This represents the video frame rate
    a = 0.5 # Tune hyperparameters
    T = 10. # Hyperparameter

    low_filtered_transforms, low_filtered_path = get_stabilized_transforms_and_trajectory(original_transforms[:,:2], original_path[:,:2], N-1, lowpass_filtering, a, T, sample_rate)
    low_filtered_frames = reconstruct_new_video(low_filtered_transforms, N-1, frames, f'./output/filtered/lowpass_frames_{a:.2f}/', f'lowpass_filtered_video_{a:.2f}.mp4', filtered=True)
    paths["Lowpass"] = low_filtered_path

    #Gaussian smoothing 
    s_g = 70 # Tune hyperparameters

    gaussian_filtered_transforms, gaussian_filtered_path = get_stabilized_transforms_and_trajectory(original_transforms, original_path, N-1, gaussian_smoothing, s_g)
    gaussian_filtered_frames = reconstruct_new_video(gaussian_filtered_transforms, N-1, frames, f'./output/filtered/gaussian_frames_{s_g:.2f}/', f'gaussian_filtered_video_{s_g:.2f}.mp4', filtered=True)
    paths["Gaussian"] = gaussian_filtered_path

    # Mean filtering 
    s_m = 150 # Tune hyperparameters

    mean_filtered_transforms, mean_filtered_path = get_stabilized_transforms_and_trajectory(original_transforms, original_path, N-1, mean_filtering, s_m)
    mean_filtered_frames = reconstruct_new_video(mean_filtered_transforms, N-1, frames, f'./output/filtered/mean_frames_{s_m:.2f}/', f'mean_filtered_video_{s_m:.2f}.mp4', filtered=True)
    paths["Mean"] = mean_filtered_path

    # Kalman filtering
    process_variance=1e-3 
    m = 1 # Tune hyperparameters

    kalman_filtered_transforms, kalman_filtered_path = get_stabilized_transforms_and_trajectory(original_transforms[:,:2], original_path[:,:2], N-1, kalman_filtering, N-1,process_variance, m)
    kalman_filtered_frames = reconstruct_new_video(kalman_filtered_transforms, N-1, frames, f'./output/filtered/kalman_frames_{m:.2f}/', f'kalman_filtered_video_{m:.2f}.mp4', filtered=True)
    paths["Kalman"] = kalman_filtered_path
    
    # Hybrid filtering 
    hybrid_filtered_transforms, hybrid_filtered_path = get_stabilized_transforms_and_trajectory(original_transforms[:,:2], gaussian_filtered_path[:,:2], N-1, kalman_filtering, N-1, 1e-3, m)  
    hybrid_filtered_frames = reconstruct_new_video(hybrid_filtered_transforms, N-1, frames, f'./output/filtered/hybrid_frames/', f'hybrid_filtered_video.mp4', filtered=True)
    paths["Hybrid"] = hybrid_filtered_path

    frames_to_video(gaussian_filtered_frames, f"./output/{keyword}/gaussian/", "gaussian_video.mp4")
    frames_to_video(mean_filtered_frames, f"./output/{keyword}/mean/", "mean_video.mp4")
    frames_to_video(kalman_filtered_frames, f"./output/{keyword}/kalman/", "kalman_video.mp4")
    frames_to_video(hybrid_filtered_frames, f"./output/{keyword}/hybrid/", "hybrid_video.mp4")

    graph_comparison_filtered_paths(original_path, paths, N-1, f"filtering_comparison_{keyword}.png")

        

def optimization_ideal(video_name, keyword):

    frames = generate_frames(video_name)

    N = len(frames)

    F_transforms, frame_shape = get_feature_transform_list_opt(frames, N)

    # Get stabilization transforms B_t by processing motion transition transforms F_t
    B_transforms = stabilize(F_transforms, frame_shape, True, None, crop_ratio=0.8)
    # Accumulate by right multiplication into C_trajectory
    # $C_{t + 1} = C_t x F_t$
    # Initialise camera trajectory as a copy since it would help with stealing index 0 and the shape of F itself
    C_trajectory = F_transforms.copy()

    for i in range(1,N):
        # Right multiply transformations to accumulate all the changes into camera trajectory
        C_trajectory[i, :, :] = C_trajectory[i - 1, :, :] @ F_transforms[i, :, :]
    # Repeat right multiplication procedure to obtain stabilized camera trajectory P
    P_trajectory = C_trajectory.copy()
    # Apply transform to C_trajectory to get P_trajectory
    for i in range(N):
        P_trajectory[i, :, :] = C_trajectory[i, :, :] @ B_transforms[i, :, :]
    
    origin = np.array([0, 0, 1])

    evolution_og = origin @ C_trajectory

    evolution_stab = origin @ P_trajectory

    graph_paths(evolution_og, evolution_stab, N, f'l1_optimization_{keyword}.png')

    new_frames = write_output(frames, N, B_transforms, (frame_shape[1], frame_shape[0]), f"./output/{keyword}/l1_optimization/", crop_ratio=0.8)

    frames_to_video(new_frames, f"./output/{keyword}/l1_optimization/", "l1_optimized_video.mp4")

optimization_ideal('./videos/youtube_vid_1.mp4', "skate")
optimization_ideal('./videos/deepstab_vid_2.mp4', "scenery")
filtered_methods_ideal('./videos/youtube_vid_1.mp4', "skate")
filtered_methods_ideal('./videos/deepstab_vid_2.mp4', "scenery")