import cv2
import numpy as np

def local_outlier_rejection(keypoint1, keypoint2, img_size, grid_size=(50, 50), threshold_distance=2.0):
    # Create a list to store filtered matches
    filtered_features = []
    filtered_matches = []

    img_grid_size_y = img_size[0] / grid_size[0]
    img_grid_size_x = img_size[1] / grid_size[1]

    # Iterate over grid cells
    for cell_y in range(grid_size[0]):
        for cell_x in range(grid_size[1]):
            # Extract features in the current grid cell
            x_lower_bound = cell_x * img_grid_size_x
            x_upper_bound = (cell_x + 1) * img_grid_size_x
            y_lower_bound = cell_y * img_grid_size_y
            y_upper_bound = (cell_y + 1) * img_grid_size_y

            cell_mask = (keypoint1[:, 0] >= y_lower_bound) & (keypoint1[:, 0] < y_upper_bound) & \
                    (keypoint1[:, 1] >= x_lower_bound) & (keypoint1[:, 1] < x_upper_bound)

            cell_features = keypoint1[cell_mask]
            cell_matches = keypoint2[cell_mask]

            if len(cell_features) >= 4:  # At least 4 points are needed for RANSAC
                # Apply RANSAC within each grid cell to estimate a translational model
                retval, inliers = cv2.estimateAffine2D(cell_features, cell_matches, method=cv2.RANSAC)

                # Retain only those matches that agree with the estimated model up to a threshold distance
                if retval is not None: 
                    residual_errors = np.linalg.norm(cell_matches - (np.column_stack((cell_features, np.ones(len(cell_features)))) @ retval.T), axis=1)
                    residual_error_mask = residual_errors < threshold_distance
                    filtered_cell_features = cell_features[residual_error_mask]
                    filtered_cell_matches = cell_matches[residual_error_mask]

                    # Store the filtered matches
                    filtered_features.extend(filtered_cell_features)
                    filtered_matches.extend(filtered_cell_matches)

    # Convert the list of filtered matches to NumPy array
    filtered_features = np.array(filtered_features)
    filtered_matches = np.array(filtered_matches)

    return filtered_features, filtered_matches


def extract_feature_keypoints_between_frames(frame1, frame2):

    # Change frames to grayscale 
    frame1_gray = cv2.cvtColor(frame1, cv2.COLOR_BGR2GRAY)
    frame2_gray = cv2.cvtColor(frame2, cv2.COLOR_BGR2GRAY)

    # Define the parameters for pyramidal Lucas-Kanade optical flow
    #lk_params = dict(winSize=(15, 15),
    #                maxLevel=2,
    #                criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03))

    # Find good features to track in the first frame
    p0 = cv2.goodFeaturesToTrack(frame1_gray, maxCorners=500, qualityLevel=0.01, minDistance=10)

    # Calculate optical flow using pyramidal Lucas-Kanade method
    if p0 is not None:
        p1, st, err = cv2.calcOpticalFlowPyrLK(frame1_gray, frame2_gray, p0, None) # , **lk_params)

        # Filter out points with no optical flow 
        p0 = p0[st == 1]
        p1 = p1[st == 1]
    else: 
        p0 = np.array([[0, 0]]) 
        p1 = np.array([[0, 0]])

    return p0, p1

def compute_feature_transform(frame1, frame2, img_shape):

    keypoint1, keypoint2 = extract_feature_keypoints_between_frames(frame1, frame2)

    final_transform,_ = cv2.estimateAffine2D(keypoint1, keypoint2)
    
    return final_transform

def compute_feature_transform_opt(frame1, frame2, img_shape):

    keypoint1, keypoint2 = extract_feature_keypoints_between_frames(frame1, frame2)

    final_transform,_ = cv2.estimateAffine2D(keypoint2, keypoint1)
    
    return final_transform

def get_feature_transform_list(frames, N):

    frame_shape = frames[0].shape 

    transforms = np.zeros((N-1, 3), np.float32)

    for i in range(N-1):
        transform = compute_feature_transform(frames[i], frames[i+1], frame_shape)

        dx = transform[0, 2]
        dy = transform[1, 2]
        dt = np.arctan2(transform[1, 0], transform[0, 0])

        # Store transformation
        transforms[i] = [dx, dy, dt]
        
    return transforms, frame_shape

def get_feature_transform_list_opt(frames, N):

    frame_shape = frames[0].shape 

    transforms = np.zeros((N, 3, 3), np.float32)
    transforms[:, :, :] = np.eye(3)

    for i in range(N-1):
        m = compute_feature_transform_opt(frames[i], frames[i+1], frame_shape)

        transforms[i+1, :, :2] = m.T
        
    return transforms, frame_shape
